from errors import CompileError
from interpreter import Machine, isconstant, isstring, isbool, isnumber
import instructions
import optimizer

EMBEDDED_PUSH_TAG = "embedded_push"

def make_embedded_push(value):
    """Returns a closure that pushed the given value onto a Machine's stack.

    We use this to embed stack pushes in the VM code, so that the interpreter
    can assume that all instructions are callable Python functions. This makes
    dispatching much faster than checking if an instruction is a constant
    (number, string, etc) or a Python function.
    """
    push = lambda vm: vm.push(value)
    push.tag = EMBEDDED_PUSH_TAG
    return push

def is_embedded_push(obj):
    """Checks if an instruction object is an embedded push."""
    return callable(obj) and hasattr(obj, "tag") and obj.tag==EMBEDDED_PUSH_TAG

def get_embedded_push_value(obj):
    """Extracts the embedded push value."""
    assert(is_embedded_push(obj))
    assert(len(obj.func_closure) == 1)
    return obj.func_closure[0].cell_contents

def check(code):
    """Checks code for obvious errors."""
    def safe_lookup(op):
        try:
            return instructions.lookup(op)
        except Exception:
            return op

    for i, a in enumerate(code):
        b = code[i+1] if i+1 < len(code) else None

        # Does instruction exist?
        if not isconstant(a):
            try:
                instructions.lookup(a)
            except KeyError, err:
                # Skip embedded push closures
                if not (len(err.args)==1 and is_embedded_push(err.args[0])):
                    raise CompileError("Instruction at index %d is unknown: %s"
                            % (i, a))

        # Invalid: <str> int
        if isstring(a) and safe_lookup(b) == instructions.cast_int:
            raise CompileError(
                "Cannot convert string to integer (index %d): %s %s" % (i, a,
                    b))

        # Invalid: <int> <binary op>
        binary_ops = [instructions.binary_not,
                      instructions.binary_or,
                      instructions.binary_and]

        if not isbool(a) and safe_lookup(b) in binary_ops:
            raise CompileError(
                "Can only use binary operators on booleans (index %d): %s %s" %
                    (i, a, b))
    return code

def compile(code, silent=True, ignore_errors=False, optimize=True):
    """Compiles subroutine-forms into a complete working code.

    A program such as:

        :sub1 <sub1 code ...> ;
        :sub2 <sub2 code ...> ;
        sub1 foo sub2 bar

    is compiled into:

        <sub1 address> call
        foo
        <sub2 address> call
        exit
        <sub1 code ...> return
        <sub2 code ...> return

    Optimizations are first done on subroutine bodies, then on the main loop
    and finally, symbols are resolved (i.e., placeholders for subroutine
    addresses are replaced with actual addresses).

    Args:
        silent: If set to False, will print optimization messages.

        ignore_errors: Only applies to the optimization engine, if set to False
            it will not raise any exceptions. The actual compilatio will still
            raise errors.

        optimize: Flag to control whether to optimize code.

    Raises:
        CompilationError - Raised if invalid code is detected.

    Returns:
        An array of code that can be run by a Machine. Typically, you want to
        pass this to a Machine without doing optimizations.

    Usage:
        source = parse("<source code>")
        code = compile(source)
        machine = Machine(code, optimize=False)
        machine.run()
    """
    assert(isinstance(code, list))

    output = []
    subroutine = {}
    builtins = Machine([]).instructions

    try:
        it = code.__iter__()
        while True:
            word = it.next()
            if word == ":":
                name = it.next()
                if name in builtins:
                    raise CompileError("Cannot shadow internal word definition '%s'." % name)
                if name in [":", ";"]:
                    raise CompileError("Invalid word name '%s'." % name)
                subroutine[name] = []
                while True:
                    op = it.next()
                    if op == ";":
                        subroutine[name].append(instructions.lookup(instructions.return_))
                        break
                    else:
                        subroutine[name].append(op)
            else:
                output.append(word)
    except StopIteration:
        pass

    # Expand all subroutine words to ["<name>", "call"]
    for name, code in subroutine.items():
        # For subroutines
        xcode = []
        for op in code:
            xcode.append(op)
            if op in subroutine:
                xcode.append(instructions.lookup(instructions.call))
        subroutine[name] = xcode

    # Compile main code (code outside of subroutines)
    xcode = []
    for op in output:
        xcode.append(op)
        if op in subroutine:
            xcode.append(instructions.lookup(instructions.call))

    # Because main code comes before subroutines, we need to explicitly add an
    # exit instruction
    output = xcode
    if len(subroutine) > 0:
        output += [instructions.lookup(instructions.exit)]

    # Optimize main code
    if optimize:
        output = optimizer.optimized(output, silent=silent, ignore_errors=False)

    # Add subroutines to output, track their locations
    location = {}
    for name, code in subroutine.items():
        location[name] = len(output)
        if optimize:
            output += optimizer.optimized(code, silent=silent, ignore_errors=False)
        else:
            output += code

    # Resolve all subroutine references
    for i, op in enumerate(output):
        if op in location:
            output[i] = location[op]

    output = native_types(output)
    if not ignore_errors:
        check(output)
    return output

def to_bool(instr):
    if isinstance(instr, bool):
        return instr
    elif instr == instructions.lookup(instructions.true_):
        return True
    elif instr == instructions.lookup(instructions.false_):
        return False
    else:
        raise CompileError("Unknown instruction: %s" % instr)

def native_types(code):
    """Convert code elements from strings to native Python types."""
    out = []
    for c in code:
        if isconstant(c, quoted=True):
            if isstring(c, quoted=True):
                v = c[1:-1]
            elif isbool(c):
                v = to_bool(c)
            elif isnumber(c):
                v = c
            else:
                raise CompileError("Unknown type %s: %s" % (type(c).__name__, c))

            # Instead of pushing constants in the code, we always push callable
            # Python functions, for fast dispatching:
            out.append(make_embedded_push(v))
        else:
            out.append(instructions.lookup(c))
    return out



class MachineError(Exception):
    """A VM runtime error."""
    pass

class ParseError(Exception):
    """An error occurring during parsing."""
    pass

class CompileError(Exception):
    """An error ocurring during compilation."""
    pass



from compiler import (check, compile)
from errors import CompileError, MachineError, ParseError
from instructions import lookup
from optimizer import constant_fold, optimized
from parser import (parse, parse_stream)
from repl import repl
from stack import Stack
from interpreter import (
    Machine,
    code_to_string,
    eval,
    execute,
    isbinary,
    isbool,
    isconstant,
    isnumber,
    isstring,
)

__version__ = "0.1.1"

__all__ = [
    "CompileError",
    "Instruction",
    "Instruction",
    "Machine",
    "MachineError",
    "ParseError",
    "Stack",
    "check",
    "code_to_string,"
    "compile",
    "constant_fold",
    "eval",
    "execute",
    "isbinary",
    "isbool",
    "isconstant",
    "isnumber",
    "isstring",
    "lookup",
    "optimized",
    "parse",
    "parse_stream",
    "repl",
]



from errors import ParseError
import StringIO
import string
import tokenize


def parse(source):
    """Parses source code returns an array of instructions suitable for
    optimization and execution by a Machine.

    Args:
        source: A string or stream containing source code.
    """
    if isinstance(source, str):
        return parse_stream(StringIO.StringIO(source))
    else:
        return parse_stream(source)

def parse_stream(stream):
    """Parse a Forth-like language and return code."""
    code = []
    tokens = tokenize.generate_tokens(stream.readline)

    def strip_whitespace(s):
        r = ""
        for ch in s:
            if not ch in string.whitespace:
                r += ch
            else:
                r += "\\x%x" % ord(ch)
        return r

    while True:
        for token, value, _, _, _ in tokens:
            if token == tokenize.NUMBER:
                try:
                    code.append(int(value))
                except ValueError, e:
                    raise ParseError(e)
            elif token in [tokenize.OP, tokenize.STRING, tokenize.NAME]:
                code.append(value)
            elif token in [tokenize.NEWLINE, tokenize.NL]:
                break
            elif token in [tokenize.COMMENT, tokenize.INDENT, tokenize.DEDENT]:
                pass
            elif token == tokenize.ENDMARKER:
                return code
            else:
                raise ParseError("Unknown token %s: '%s'" %
                        (tokenize.tok_name[token], strip_whitespace(value)))
    return code




from errors import MachineError

class Stack(object):
    """A stack of values."""
    def __init__(self, values=None):
        self._values = values

        if self._values is None:
            self._values = []

    def pop(self):
        if len(self._values) == 0:
            raise MachineError("Stack underflow")
        return self._values.pop()

    def push(self, value):
        self._values.append(value)

    @property
    def top(self):
        return None if len(self._values) == 0 else self._values[-1]

    def __str__(self):
        return str(self._values)

    def __repr__(self):
        return "<Stack: values=%s>" % self._values

    def __len__(self):
        return len(self._values)

    def __getitem__(self, key):
        return self._values[key]

    def __eq__(self, obj):
        return self._values == obj._values

    def __ne__(self, obj):
        return self._values != obj._values



"""
Contains a collection of useful instructions, or machine primivites.
"""

import errors
import interpreter

def _assert_number(*args):
    for arg in args:
        if not interpreter.isnumber(arg):
            raise errors.MachineError("Not an integer: %s" % str(arg))

def _assert_bool(*args):
    for arg in args:
        if not interpreter.isbool(arg):
            raise errors.MachineError("Not a boolean: %s" % str(arg))

def _assert_binary(*args):
    for arg in args:
        if not interpreter.isbinary(arg):
            raise errors.MachineError("Not boolean or numerical: %s" % str(arg))

def add(vm):
    a = vm.pop()
    b = vm.pop()
    _assert_number(a, b)
    vm.push(a + b)

def sub(vm):
    a = vm.pop()
    b = vm.pop()
    _assert_number(a, b)
    vm.push(b - a)

def call(vm):
    vm.return_stack.push(vm.instruction_pointer)
    jmp(vm)

def return_(vm):
    vm.instruction_pointer = vm.return_stack.pop()

def mul(vm, modulus=None):
    a = vm.pop()
    b = vm.pop()
    _assert_number(a, b)

    if modulus is None:
        vm.push(a * b)
    else:
        vm.push((a * b) % modulus)

def div(vm):
    divisor = vm.pop()
    dividend = vm.pop()
    _assert_number(dividend, divisor)
    if divisor == 0:
        raise errors.MachineError(ZeroDivisionError("Division by zero"))
    vm.push(dividend / divisor)

def mod(vm):
    a = vm.pop()
    b = vm.pop()
    _assert_number(a, b)
    if a == 0:
        raise errors.MachineError(ZeroDivisionError("Division by zero"))
    vm.push(b % a)

def abs_(vm):
    _assert_binary(vm.top)
    vm.push(abs(vm.pop()))

def nop(vm):
    """No operation; do nothing."""
    pass

def exit(vm):
    raise StopIteration

def dup(vm):
    vm.push(vm.top)

def over(vm):
    b = vm.pop()
    a = vm.pop()
    vm.push(a)
    vm.push(b)
    vm.push(a)

def drop(vm):
    vm.pop()

def swap(vm):
    b = vm.pop()
    a = vm.pop()
    vm.push(b)
    vm.push(a)

def rot(vm):
    """Rotate topmost three items once to the left. ( a b c -- b c a )"""
    c = vm.pop()
    b = vm.pop()
    a = vm.pop()
    vm.push(b)
    vm.push(c)
    vm.push(a)

def r_at(vm):
    """Pop top of return stack and push onto data stack."""
    vm.push(vm.data_stack.pop())

def r_gt(vm):
    """Copy top of the return stack and push onto the data stack."""
    vm.push(vm.data_stack.top)

def write(vm, flush=True):
    value = str(vm.pop())
    try:
        if vm.output is not None:
            vm.output.write(value)
            if flush:
                vm.output.flush()
    except IOError:
        raise errors.MachineError(StopIteration)

def at(vm):
    """Pushes previous position to return stack."""
    vm.return_stack.push(vm.instruction_pointer - 1)

def dot(vm, flush=True):
    write(vm, flush=False)
    try:
        if vm.output is not None:
            vm.output.write("\n")
            if flush:
                vm.output.flush()
    except IOError:
        raise errors.MachineError(IOError)

def read(vm):
    line = vm.input.readline().rstrip()
    vm.push(line)

    # To be compatible with the old code that relied on raw_input raising
    # EOFError, we should also do it for a while (later on, we should use
    # another mechanism).
    if line == "":
        raise EOFError()

def cast_int(vm):
    try:
        int(vm.top)
    except ValueError:
        raise errors.MachineError("Cannot be cast to int: '%s'" % str(vm.top))
    else:
        vm.push(int(vm.pop()))

def cast_bool(vm):
    try:
        bool(vm.top)
    except ValueError:
        raise errors.MachineError("Cannot be cast to bool: '%s'" % str(vm.top))
    else:
        vm.push(bool(vm.pop()))

def cast_str(vm):
    try:
        vm.push(str(vm.pop()))
    except ValueError, e:
        raise errors.MachineError(e)

def equal(vm):
    vm.push(vm.pop() == vm.pop())

def not_equal(vm):
    vm.push(vm.pop() != vm.pop())

def less(vm):
    a = vm.pop()
    vm.push(a < vm.pop())

def greater(vm):
    a = vm.pop()
    vm.push(a > vm.pop())

def true_(vm):
    vm.push(True)

def false_(vm):
    vm.push(False)

def if_stmt(vm):
    false_clause = vm.pop()
    true_clause = vm.pop()
    test = vm.pop()

    # False values: False, 0, "", everyting else is true
    if interpreter.isbool(test) and test == False:
        result = False
    elif interpreter.isstring(test) and len(test) == 0:
        result = False
    elif interpreter.isnumber(test) and test == 0:
        result = False
    else:
        result = True

    if result == True:
        vm.push(true_clause)
    else:
        vm.push(false_clause)

def jmp(vm):
    if not (isinstance(vm.top, int) or isinstance(vm.top, long)):
        raise errors.MachineError("Jump address must be an integer: %s" %
                str(vm.top))
    addr = vm.pop()
    if 0 <= addr < len(vm.code):
        vm.instruction_pointer = addr
    else:
        raise errors.MachineError("Jump address out of range: %s" % str(addr))

def dump_stack(vm):
    vm.output.write("Data stack:\n")
    for v in reversed(vm.data_stack._values):
        vm.output.write(" - type %s, value '%s'\n" % (type(v), v))

    vm.output.write("Return stack:\n")
    for v in reversed(vm.return_stack._values):
        vm.output.write(" - address %s\n" % str(v))

    vm.output.flush()

def bitwise_and(vm):
    a = vm.pop()
    b = vm.pop()
    _assert_binary(a, b)
    vm.push(b & a)

def bitwise_or(vm):
    a = vm.pop()
    b = vm.pop()
    _assert_binary(a, b)
    vm.push(b | a)

def bitwise_xor(vm):
    a = vm.pop()
    b = vm.pop()
    _assert_binary(a, b)
    vm.push(b ^ a)

def bitwise_complement(vm):
    a = vm.pop()
    _assert_binary(a)
    vm.push(~a)

def binary_not(vm):
    a = vm.pop()
    _assert_bool(a)
    vm.push(not a)

def binary_and(vm):
    a = vm.pop()
    b = vm.pop()
    _assert_bool(a, b)
    vm.push(b and a)

def binary_or(vm):
    a = vm.pop()
    b = vm.pop()
    _assert_bool(a, b)
    vm.push(b or a)

def negate(vm):
    _assert_number(vm.top)
    vm.push(-vm.pop())

def lookup(instruction, instructions = None):
    """Looks up instruction, which can either be a function or a string.
    If it's a string, returns the corresponding method.
    If it's a function, returns the corresponding name.
    """
    if instructions is None:
        instructions = default_instructions

    if isinstance(instruction, str):
        return instructions[instruction]
    elif hasattr(instruction, "__call__"):
        rev = dict(((v,k) for (k,v) in instructions.items()))
        return rev[instruction]
    else:
        raise errors.MachineError(KeyError("Unknown instruction: %s" %
            str(instruction)))

default_instructions = {
    "%":      mod,
    "&":      bitwise_and,
    "*":      mul,
    "+":      add,
    "-":      sub,
    ".":      dot,
    "/":      div,
    "<":      less,
    "<>":     not_equal,
    "=":      equal,
    ">":      greater,
    "@":      at,
    "^":      bitwise_xor,
    "abs":    abs_,
    "and":    binary_and,
    "bool":   cast_bool,
    "call":   call,
    "drop":   drop,
    "dup":    dup,
    "exit":   exit,
    "false":  false_,
    "if":     if_stmt,
    "int":    cast_int,
    "jmp":    jmp,
    "negate": negate,
    "nop":    nop,
    "not":    binary_not,
    "or":     binary_or,
    "over":   over,
    "read":   read,
    "return": return_,
    "rot":    rot,
    "str":    cast_str,
    "swap":   swap,
    "true":   true_,
    "write":  write,
    "|":      bitwise_or,
    "~":      bitwise_complement,
    #".stack": dump_stack, # Parser doesn't like this
    #"r>":     r_gt, # Parser doesn't like this
    #"r@":     r_at, # Parser doesn't like this
}




import compiler
import errors
import instructions
import parser
import stack
import sys

def code_to_string(code):
    s = []
    for op in code:
        if isconstant(op):
            if isstring(op):
                s.append(repr(op))
            else:
                s.append(str(op))
        elif compiler.is_embedded_push(op):
            v = compiler.get_embedded_push_value(op)
            s.append('"%s"' % repr(v)[1:-1] if isinstance(v, str) else repr(v))
        else:
            s.append(str(instructions.lookup(op)))
    return " ".join(s)

def isstring(args, quoted=False):
    """Checks if value is a (quoted) string."""
    isquoted = lambda c: c[0]==c[-1] and c[0] in ['"', "'"]

    if quoted:
        check = lambda c: isinstance(c, str) and isquoted(c)
    else:
        check = lambda c: isinstance(c, str)

    if isinstance(args, list):
        return all(map(check, args))
    else:
        return check(args)

def isnumber(*args):
    """Checks if value is an integer, long integer or float.

    NOTE: Treats booleans as numbers, where True=1 and False=0.
    """
    return all(map(lambda c: isinstance(c, int) or isinstance(c, long) or
        isinstance(c, float), args))

def isbool(*args):
    """Checks if value is boolean."""
    true_or_false = [instructions.lookup(instructions.true_),
                     instructions.lookup(instructions.false_)]
    return all(map(lambda c: isinstance(c, bool) or c in true_or_false, args))

def isbinary(*args):
    """Checks if value can be part of binary/bitwise operations."""
    return all(map(lambda c: isnumber(c) or isbool(c), args))

def isconstant(args, quoted=False):
    """Checks if value is a boolean, number or string."""
    check = lambda c: isbool(c) or isnumber(c) or isstring(c, quoted)
    if isinstance(args, list):
        return all(map(check, args))
    else:
        return check(args)

def execute(source, optimize=True, output=sys.stdout, input=sys.stdin, steps=-1):
    """Compiles and runs program, returning the machine used to execute the
    code.

    Args:
        optimize: Whether to optimize the code after parsing it.
        output: Stream which program can write output to.
        input: Stream which program can read input from.
        steps: An optional maximum number of instructions to execute on the
            virtual machine.  Set to -1 for no limit.

    Returns:
        A Machine instance.
    """
    code = compiler.compile(parser.parse(source), optimize=optimize)
    machine = Machine(code, output=output, input=input)
    return machine.run(steps)

def eval(source, optimize=True, output=sys.stdout, input=sys.stdin, steps=-1):
    """Compiles and runs program, returning the values on the stack.

    To return the machine instead, see execute().

    Args:
        optimize: Whether to optimize the code after parsing it.
        output: Stream which program can write output to.
        input: Stream which program can read input from.
        steps: An optional maximum number of instructions to execute on the
            virtual machine.  Set to -1 for no limit.

    Returns:
        None: If the stack is empty
        obj: If the stack contains a single value
        [obj, obj, ...]: If the stack contains many values
    """
    machine = execute(source, optimize=optimize, output=output, input=input,
            steps=steps)
    ds = machine.stack

    if len(ds) == 0:
        return None
    elif len(ds) == 1:
        return ds[-1]
    else:
        return ds


class Machine(object):
    """A virtual machine with code, a data stack and an instruction stack."""

    def __init__(self, code, output=sys.stdout, input=sys.stdin):
        """
        Args:
            code: The code to run.
            output: Output stream that the machine's code can write to.
            input: Input stream that the machine's code can read from.
            optimize: If True, optimize the given code.
        """
        self.reset()
        self.code = code
        self.output = output
        self.input = input
        self.instructions = instructions.default_instructions

    def lookup(self, instruction):
        """Looks up name-to-function or function-to-name."""
        return instructions.lookup(instruction, self.instructions)

    @property
    def stack(self):
        """Returns the data (operand) stack values."""
        return self.data_stack._values

    def reset(self):
        """Reset stacks and instruction pointer."""
        self.data_stack = stack.Stack()
        self.return_stack = stack.Stack()
        self.instruction_pointer = 0
        return self

    def __repr__(self):
        return "<Machine: ip=%d |ds|=%d |rs|=%d top=%s>" % (self.instruction_pointer,
                len(self.data_stack), len(self.return_stack), str(self.top))

    def __str__(self):
        return self.__repr__()

    @property
    def code_string(self):
        """Returns code as a parseable string."""
        return code_to_string(self.code)

    def pop(self):
        """Pops the data stack, returning the value."""
        try:
            return self.data_stack.pop()
        except errors.MachineError, e:
            raise errors.MachineError("%s: At index %d in code: %s" %
                    (e, self.instruction_pointer, self.code_string))

    def push(self, value):
        """Pushes a value on the data stack."""
        self.data_stack.push(value)

    @property
    def top(self):
        """Returns the top of the data stack."""
        return self.data_stack.top

    def step(self):
        """Executes one instruction and stops."""
        op = self.code[self.instruction_pointer]
        self.instruction_pointer += 1
        op(self)

    def run(self, steps=None):
        """Run threaded code in machine.

        Args:
            steps: If specified, run that many number of instructions before
            stopping.
        """
        try:
            while self.instruction_pointer < len(self.code):
                self.step()

                if steps is not None:
                    steps -= 1
                    if steps == 0:
                        break
        except StopIteration:
            pass
        except EOFError:
            pass
        return self



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name="crianza",
    packages=["crianza"],
    version="0.1.1",
    description="Simple Forth-like VM and genetic programming framework.",
    author="Christian Stigen Larsen",
    author_email="csl@csl.name",
    url="https://github.com/cslarsen/crianza",
    download_url="https://github.com/cslarsen/crianza/tarball/0.1.1",
    scripts=["bin/crianza"],
    license="LICENSE.txt",
    long_description=open("README.rst").read(),
    zip_safe=True,
    keywords=["vm", "virtual machine", "genetic programming", "interpreter"],
    classifiers=[],
)



import errors
import instructions
import interpreter


def optimized(code, silent=True, ignore_errors=True):
    """Performs optimizations on already parsed code."""
    return constant_fold(code, silent=silent, ignore_errors=ignore_errors)

def constant_fold(code, silent=True, ignore_errors=True):
    """Constant-folds simple expressions like 2 3 + to 5.

    Args:
        code: Code in non-native types.
        silent: Flag that controls whether to print optimizations made.
        ignore_errors: Whether to raise exceptions on found errors.
    """
    # Loop until we haven't done any optimizations.  E.g., "2 3 + 5 *" will be
    # optimized to "5 5 *" and in the next iteration to 25.  Yes, this is
    # extremely slow, big-O wise. We'll fix that some other time. (TODO)

    arithmetic = map(instructions.lookup, [
        instructions.add,
        instructions.bitwise_and,
        instructions.bitwise_or,
        instructions.bitwise_xor,
        instructions.div,
        instructions.equal,
        instructions.greater,
        instructions.less,
        instructions.mod,
        instructions.mul,
        instructions.sub,
    ])

    divzero = map(instructions.lookup, [
        instructions.div,
        instructions.mod,
    ])

    lookup = instructions.lookup

    def isfunction(op):
        try:
            instructions.lookup(op)
            return True
        except KeyError:
            return False

    def isconstant(op):
        return op is None or interpreter.isconstant(op, quoted=True) or not isfunction(op)

    keep_running = True
    while keep_running:
        keep_running = False
        # Find two consecutive numbes and an arithmetic operator
        for i, a in enumerate(code):
            b = code[i+1] if i+1 < len(code) else None
            c = code[i+2] if i+2 < len(code) else None

            #print("-- a=%s b=%s c=%s lookup=%s" % (a,b,c,lookup(c)))

            # Constant fold arithmetic operations (TODO: Move to check-func)
            if interpreter.isnumber(a, b) and c in arithmetic:
                # Although we can detect division by zero at compile time, we
                # don't report it here, because the surrounding system doesn't
                # handle that very well. So just leave it for now.  (NOTE: If
                # we had an "error" instruction, we could actually transform
                # the expression to an error, or exit instruction perhaps)
                if b==0 and c in divzero:
                    if ignore_errors:
                        continue
                    else:
                        raise errors.CompileError(ZeroDivisionError(
                            "Division by zero"))

                # Calculate result by running on a machine (lambda vm: ... is
                # embedded pushes, see compiler)
                result = interpreter.Machine([lambda vm: vm.push(a), lambda vm:
                    vm.push(b), instructions.lookup(c)]).run().top
                del code[i:i+3]
                code.insert(i, result)

                if not silent:
                    print("Optimizer: Constant-folded %d %d %s to %d" % (a,b,c,result))

                keep_running = True
                break

            # Translate <constant> dup to <constant> <constant>
            if isconstant(a) and b == lookup(instructions.dup):
                code[i+1] = a
                if not silent:
                    print("Optimizer: Translated %s %s to %s %s" % (a,b,a,a))
                keep_running = True
                break

            # Dead code removal: <constant> drop
            if isconstant(a) and b == lookup(instructions.drop):
                del code[i:i+2]
                if not silent:
                    print("Optimizer: Removed dead code %s %s" % (a,b))
                keep_running = True
                break

            if a == lookup(instructions.nop):
                del code[i]
                if not silent:
                    print("Optimizer: Removed dead code %s" % a)
                keep_running = True
                break

            # Dead code removal: <integer> cast_int
            if isinstance(a, int) and b == lookup(instructions.cast_int):
                del code[i+1]
                if not silent:
                    print("Optimizer: Translated %s %s to %s" % (a,b,a))
                keep_running = True
                break

            # Dead code removal: <string> cast_str
            if isinstance(a, str) and b == lookup(instructions.cast_str):
                del code[i+1]
                if not silent:
                    print("Optimizer: Translated %s %s to %s" % (a,b,a))
                keep_running = True
                break

            # Dead code removal: <boolean> cast_bool
            if isinstance(a, bool) and b == lookup(instructions.cast_bool):
                del code[i+1]
                if not silent:
                    print("Optimizer: Translated %s %s to %s" % (a,b,a))
                keep_running = True
                break

            # <c1> <c2> swap -> <c2> <c1>
            if isconstant(a) and isconstant(b) and c == lookup(instructions.swap):
                del code[i:i+3]
                code = code[:i] + [b, a] + code[i:]
                if not silent:
                    print("Optimizer: Translated %s %s %s to %s %s" %
                            (a,b,c,b,a))
                keep_running = True
                break

            # a b over -> a b a
            if isconstant(a) and isconstant(b) and c == lookup(instructions.over):
                code[i+2] = a
                if not silent:
                    print("Optimizer: Translated %s %s %s to %s %s %s" %
                            (a,b,c,a,b,a))
                keep_running = True
                break

            # "123" cast_int -> 123
            if interpreter.isstring(a) and b == lookup(instructions.cast_int):
                try:
                    number = int(a)
                    del code[i:i+2]
                    code.insert(i, number)
                    if not silent:
                        print("Optimizer: Translated %s %s to %s" % (a, b,
                            number))
                    keep_running = True
                    break
                except ValueError:
                    pass

            if isconstant(a) and b == lookup(instructions.cast_str):
                del code[i:i+2]
                code.insert(i, str(a))
                if not silent:
                    print("Optimizer: Translated %s %s to %s" % (a, b, str(a)))
                keep_running = True
                break

            if isconstant(a) and b == lookup(instructions.cast_bool):
                del code[i:i+2]
                code.insert(i, bool(a))
                if not silent:
                    print("Optimizer: Translated %s %s to %s" % (a, b, bool(a)))
                keep_running = True
                break
    return code



"""
Simple genetic programming using the virtual machine.

TIP: Run with pypy for speed.

See examples in examples-genetic/
"""

import crianza
import random
import sys

def _log(s, stream=sys.stdout):
    stream.write(s)
    stream.flush()

def tanimoto_coefficient(a, b):
    """Measured similarity between two points in a multi-dimensional space.

    Returns:
        1.0 if the two points completely overlap,
        0.0 if the two points are infinitely far apart.
    """
    return sum(map(lambda (x,y): float(x)*float(y), zip(a,b))) / sum([
          -sum(map(lambda (x,y): float(x)*float(y), zip(a,b))),
           sum(map(lambda x: float(x)**2, a)),
           sum(map(lambda x: float(x)**2, b))])

def weighted_tanimoto(a, b, weights):
    """Same as the Tanimoto coefficient, but wit weights for each dimension."""
    weighted = lambda s: map(lambda (x,y): float(x)*float(y), zip(s, weights))
    return tanimoto_coefficient(weighted(a), weighted(b))

def average(sequence, key):
    """Averages a sequence based on a key."""
    return sum(map(key, sequence)) / float(len(sequence))

def weighted_choice(choices):
    # Taken from http://stackoverflow.com/a/3679747/21028
    total = sum(w for c, w in choices)
    r = random.uniform(0, total)
    upto = 0
    for c, w in choices:
        if upto + w > r:
            return c
        upto += w
    assert False, "Shouldn't get here"

def stochastic_choice(machines):
    """Stochastically choose one random machine distributed along their
    scores."""
    if len(machines) < 4:
        return random.choice(machines)

    r = random.random()
    if r < 0.5:
        return random.choice(machines[:len(machines)/4])
    elif r < 0.75:
        return random.choice(machines[:len(machines)/2])
    else:
        return random.choice(machines)

    #This is better, but waaaay too slow:
    # w = [(1.0 - m.score()) for m in machines]
    # return weighted_choice(zip(machines, w))

def randomize(vm,
        length=(10,10),
        ints=(0,999),
        strs=(1,10),
        chars=(32,126),
        instruction_ratio=0.5,
        number_string_ratio=0.8,
        exclude=map(crianza.instructions.lookup, [".", "exit", "read", "write", "str"]),
        restrict_to=None):

    """Replaces existing code with completely random instructions. Does not
    optimize code after generating it.

    Args:
        length: Tuple of minimum and maximum code lengths. Code length will
        be a random number between these two, inclusive values.

        ints: Integers in the code will be selected at random from this
        inclusive range.

        strs: Inclusive range of the length of strings in the code.

        chars: Inclusive range of characters in random strings.

        instruction_ratio: Ratio of instructions to numbers/strings,
        meaning that if this value is 0.5 then there will just as many
        instructions in the code as there are numbers and strings.

        number_string_ratio: Ratio of numbers to strings.

        exclude: Excluded instructions. For genetic programming, one wants
        to avoid the program to hang for user input.  The default value is
        to exclude console i/o and debug instructions.

        restrict_to: Limit instructions to the given list.

    Returns:
        The VM.
    """
    vm.code = []
    instructions = set(vm.instructions.values()) - set(exclude)

    if restrict_to is not None:
        instructions = instructions.intersection(set(restrict_to))

    instructions = list(instructions)

    for _ in xrange(random.randint(*length)):
        r = random.random()
        if r <= instruction_ratio:
            # Generate a random instruction
            vm.code.append(random.choice(instructions))
        elif r <= number_string_ratio:
            # Generate a random number
            vm.code.append(crianza.compiler.make_embedded_push(random.randint(*ints)))
        else:
            # Generate a random string
            vm.code.append(crianza.compiler.make_embedded_push('%s' %
                "".join(chr(random.randint(*chars)) for n in xrange(0,
                    random.randint(*strs)))))
    return vm

def crossover(m, f):
    """Produces an offspring from two Machines, whose code is a
    combination of the two."""
    i = random.randint(0, len(m.code))
    j = random.randint(0, len(f.code))
    return m.code[:i] + f.code[j:]


class GeneticMachine(crianza.Machine):
    def __init__(self, code):
        super(GeneticMachine, self).__init__(code)
        self._error = False

    def setUp(self):
        """Called before each invocation of run()."""
        self.reset()
        return self

    def tearDown(self):
        """Called after each invocation of run()."""
        return self

    def new(self, *args, **kw):
        """A virtual constructor.

        Subclasses should override this method to construct correct instances.
        """
        return GeneticMachine(*args, **kw)

    def randomize(self, **kw):
        """Creates a new random code."""
        return randomize(self, **kw)

    def crossover(self, other):
        """Produce offspring from this and another instance."""
        return self.new(crossover(self, other))

    def mutate(self):
        """Mutates code."""
        # Choose a random position
        if len(self.code) == 0:
            return

        index = random.randint(0, len(self.code)-1)
        mutation_type = random.random()

        if mutation_type < 0.5:
            # Change
            self.code[index] = self.new().randomize().code[0]
        elif mutation_type < 0.75:
            # Deletion
            del self.code[index]
        else:
            # Insertion
            self.code.insert(index, self.new().randomize().code[0])

    def run(self, steps=10):
        """Executes up to `steps` instructions."""
        try:
            super(GeneticMachine, self).run(steps)
            self._error = False
        except StopIteration:
            self._error = False
        except Exception:
            self._error = True

    def score(self):
        """Returns a machine's fitness as a number from 0.0 (perfect score) to
        1.0 (worst score).

        This method must be overriden to produce selection pressure, i.e. to
        direct what kind of problem this machine should solve.
        """
        return 1.0 if self._error else 0.0

    def stop(iterations, generation):
        """Conditions for when to stop the simulation."""
        return iterations >= 10000


def iterate(MachineClass, stop_function=lambda iterations: iterations < 10000,
        machines=1000, survival_rate=0.05, mutation_rate=0.075, silent=False):
    """Creates a bunch of machines, runs them for a number of steps and then
    gives them a fitness score.  The best produce offspring that are passed on
    to the next generation.

    Args:
        setup: Function taking a Machine argument, makes the machine ready
        machine_class: A class that should inherit from Machine and contain the
        following methods:
            setUp(self) [optional]: Run before each new run, should at least
            call self.reset() for each run.

            tearDown(self) [optional]: Run after each run.

            crossover(self, other): Returns a Machine offspring between itself and
            another Machine.

            score(self):  Score the machine. Lower is better. Can return a
            tuple of values used to sort the machines from best to worst in
            fitness.

        stop_function: An optional function that takes the current generation
        number and returns True if the processing should stop.

        machines: The number of machines to create for each generation.

        steps: The number of instructions each machine is allowed to execute in
        each run.

        code_length: Tuple of (minimum code length, maximum code length) for a
        Machine.

        survival_ratio: Ratio of the best Machines which are allowed to produce
        offspring for the next generation.

        mutation_rate: Rate for each machine's chance of being mutated.
    """
    def make_random(n):
        return MachineClass().randomize()

    def run_once(m):
        m.setUp()
        m.run()
        m.tearDown()
        return m

    def make_offspring(survivors):
        a = stochastic_choice(survivors)
        b = stochastic_choice(survivors)
        return a.crossover(b)

    generation = map(make_random, xrange(machines))
    survivors = generation

    if silent:
        log = lambda s,stream=None: None
    else:
        log = _log

    try:
        iterations = 0
        while not stop_function(iterations, survivors):
            iterations += 1
            log("running ... ")

            # Run all machines in this generation
            generation = map(run_once, generation)

            # Sort machines from best to worst
            generation = sorted(generation, key=lambda m: m.score())

            # Select the best
            survivors = generation[:int(survival_rate * len(generation))]

            # Remove code larger than 50
            for s in survivors:
                if len(s.code) >= 50:
                    s.code = s.code[:50]

            # Remove dead ones
            survivors = [s for s in survivors if len(s.code)>0]

            #survivors = [s for s in survivors if len(s.code)<=50]
            # All dead? start with a new set
            if len(survivors) == 0:
                log("\nNo survivors, restarting")
                survivors = map(make_random, xrange(machines))
                generation = survivors
                continue

            # Create a new generation based on the survivors.
            log("crossover ... ")
            cross = lambda _: make_offspring(survivors)
            generation = map(cross, xrange(machines))

            # Add mutations from time to time
            for m in generation:
                if random.random() > mutation_rate:
                    m.mutate()

            log("\rgen %d 1-fitness %.12f avg code len %.2f avg stack len %.2f\n" %
                (iterations,
                 average(survivors, lambda m: m.score()),
                 average(survivors, lambda m: len(m.code)),
                 average(survivors, lambda m: len(m.stack) + len(m.return_stack))))
    except KeyboardInterrupt:
        pass

    return survivors



import sys

from compiler import compile, is_embedded_push, get_embedded_push_value
from errors import ParseError, MachineError, CompileError
from parser import parse
from interpreter import isstring, Machine


def repl(optimize=True, persist=True):
    """Starts a simple REPL for this machine.

    Args:
        optimize: Controls whether to run inputted code through the
        optimizer.

        persist: If True, the machine is not deleted after each line.
    """

    def print_code(vm, ops_per_line=8):
        """Prints code and state for VM."""
        print("IP: %d" % vm.instruction_pointer)
        print("DS: %s" % str(vm.stack))
        print("RS: %s" % str(vm.return_stack))

        def to_str(op):
            if is_embedded_push(op):
                op = get_embedded_push_value(op)

            if isstring(op):
                return repr(op)[1:-1]
            elif callable(op):
                return vm.lookup(op)
            else:
                return str(op)

        for addr, op in enumerate(vm.code):
            if (addr % ops_per_line) == 0 and (addr+1) < len(vm.code):
                if addr > 0:
                    sys.stdout.write("\n")
                sys.stdout.write("%0*d  " % (max(4, len(str(len(vm.code)))), addr))
            sys.stdout.write("%s " % to_str(op))
        sys.stdout.write("\n")

    print("Extra commands for the REPL:")
    print(".code    - print code")
    print(".raw     - print raw code")
    print(".quit    - exit immediately")
    print(".reset   - reset machine (IP and stacks)")
    print(".restart - create a clean, new machine")
    print(".clear   - same as .restart")
    print(".stack   - print data stack")
    print("")

    machine = Machine([])

    def match(s, *args):
        return any(map(lambda arg: s.strip()==arg, args))

    while True:
        try:
            source = raw_input("> ").strip()

            if source[0] == ".":
                if match(source, ".quit"):
                    return
                elif match(source, ".code"):
                    print_code(machine)
                elif match(source, ".raw"):
                    print(machine.code)
                elif match(source, ".reset"):
                    machine.reset()
                elif match(source, ".restart", ".clear"):
                    machine = Machine([])
                elif match(source, ".stack"):
                    print(machine.stack)
                else:
                    raise ParseError("Unknown command: %s" % source)
                continue

            code = compile(parse(source), silent=False, optimize=optimize)

            if not persist:
                machine.reset()

            machine.code += code
            machine.run()
        except EOFError:
            return
        except KeyboardInterrupt:
            return
        except ParseError, e:
            print("Parser error: %s" % e)
        except MachineError, e:
            print("Machine error: %s" % e)
        except CompileError, e:
            print("Compile error: %s" % e)
